import { world, system, TimeOfDay } from "@minecraft/server";
import * as Vector3 from "./vector3";

let scoreboard = world.scoreboard;
export const overworld = world.getDimension('overworld');
export const nether = world.getDimension('nether');
export const the_end = world.getDimension('the_end');
export const dimensions = [overworld, nether, the_end];

export let player0 = undefined;

export function randomInt(min, max) {
	return Math.floor(Math.random() * (max - min + 1))+min;
}
export function floorVector(vector)
{
	return { x: Math.floor(vector.x), y: Math.floor(vector.y+0.5), z: Math.floor(vector.z) };
}

export async function waitForTicks(ticks, func) {
    return await system.runTimeout(func, ticks);
}

export async function waitForLoopEvent(ticks, callback) {
    return new Promise(resolve => {
        let id = 0;
        const event = async function () {
            var _a;
            if (await callback()) {
                resolve();
            }
            else {
                id = system.runTimeout(event, ticks);
            }
        };
        system.run(event);
    });
}

export async function onLoad() {
	if (player0)
		return;
	
	await waitForLoopEvent(1, () => {
		return world.getAllPlayers().length > 0;
	});
	player0 = await world.getAllPlayers()[0];
	if (!player0.hasTag("goe_bos_player0"))
		player0.addTag("goe_bos_player0");
}


export async function runCommand(command, dim) {
	if (dim === undefined)
		return await overworld.runCommandAsync(command);
	return await dim.runCommandAsync(command);
}

export async function runPlayer0Command(command) {
	return await player0.runCommandAsync(command);
}

export async function runPlayerCommand(entity, command) {
	return await entity.runCommandAsync(command);
}
export async function runDelayedPlayerCommand(delay, entity, command) {
	return await system.runTimeout(() => { runPlayerCommand(entity, command); }, delay);
}
export async function runDelayedEntityCommand(delay, entity, command) {
	return await system.runTimeout(() => { runEntityCommand(entity, command); }, delay);
}
export async function runMobCommand(entity, command) {
	return await entity.runCommandAsync(command);
}
export async function runEntityCommand(entity, command) {
	return await entity.runCommandAsync(command);
}

export async function getAllMobs()
{
	let entities = await dimensions.flatMap((dim) => (dim.getEntities()));
	entities = entities.filter( (entity) => entity.isValid() && entity.typeId !== "minecraft:player");
	return entities;
}
export async function getEntities(entityType)
{
	const entities = await dimensions.flatMap((dim) => (dim.getEntities({ type: entityType })));
	return entities;
}
export async function getAllEntitiesWithTags(entityTags)
{
	const entities = await dimensions.flatMap((dim) => (dim.getEntities({ tags: entityTags })));
	return entities;
}
export async function getEntitiesWithTags(entityType, entityTags)
{
	const entities = await dimensions.flatMap((dim) => (dim.getEntities({ type: entityType, tags: entityTags })));
	return entities;
}
export async function getEntityNearEntity(entityType, nearbyEntity)
{
	const entities = await nearbyEntity.dimension.getEntities( { type: entityType, location: nearbyEntity.location, closest: 1 } );
	return entities[0];
}
export async function getEntityInLocation(entityType, dim, centerLocation)
{
	const entities = await dim.getEntities( { type: entityType, location: centerLocation, closest: 1 } );
	return entities;
}
export async function getPlayerInLocation(dim, centerLocation)
{
	const players = await dim.getPlayers( { location: centerLocation, closest: 1 } );
	return players;
}
export async function getPlayersInRange(dim, centerLocation, minDist, maxDist)
{
	const players = await dim.getPlayers( { location: centerLocation, minDistance: minDist, maxDistance: maxDist } );
	return players;
}
export async function getPlayersWithTags(playerTags)
{
	const players = await dimensions.flatMap((dim) => (dim.getPlayers({ tags: playerTags })));
	return players;
}
export async function getEntitiesInRange(entityType, dim, centerLocation, minDist, maxDist)
{
	const entities = await dim.getEntities( { type: entityType, location: centerLocation, minDistance: minDist, maxDistance: maxDist } );
	return entities;
}

export async function getEntitiesInArea(entityType, dim, corner1, corner2, checkDimensions, tag)
{
	// checkDimensions - what dimensions to check - "x", "xz", "xyz", etc
	let entities = await dim.getEntities( { type: entityType } );
	if (tag !== undefined)
		entities = entities.filter( (entity) =>  entity.hasTag(tag));
	return entities.filter( (entity) =>  Vector3.locationIsInArea(entity.location, corner1, corner2, checkDimensions));
}
export async function getPlayersInArea(dim, corner1, corner2, checkDimensions, tag)
{
	// checkDimensions - what dimensions to check - "x", "xz", "xyz", etc
	let players = await dim.getPlayers();
	if (tag !== undefined)
		players = players.filter( (player) =>  player.hasTag(tag));
	return players.filter( (player) =>  Vector3.locationIsInArea(player.location, corner1, corner2, checkDimensions));
}

export async function getAllEntitiesInRange(dim, centerLocation, minDist, maxDist)
{
	const entities = await dim.getEntities( { location: centerLocation, minDistance: minDist, maxDistance: maxDist } );
	return entities;
}

export async function countItemInInventory(player, itemTypeId)
{
	const inventory = player.getComponent("minecraft:inventory").container;
	let itemCount = 0;
	for (let i = 0; i < inventory.size; i++) {
		let item = inventory.getItem(i);
		if (item !== undefined && item.typeId === itemTypeId)
			itemCount += item.amount;
	}
	return itemCount
}

const DEFAULT_MESSAGE_COOLDOWN = 3; // Cooldown time in seconds

// Add cooldown (in seconds) to stop K/D actionbar overriding message
export async function setMessageCooldown(cooldown = 1) {
	world.setDynamicProperty("goe_bos_message_cooldown", cooldown);
}

export async function showPlayersErrorMessage(entity, target, message)
{
	await setMessageCooldown(DEFAULT_MESSAGE_COOLDOWN);
	await actionbar(entity, target, message);
	await entity.runCommandAsync(`playsound goe_bos:error @a[r=5] ~ ~ ~ 0.8`);
}
export async function showPlayerErrorMessage(player, message)
{
	await setMessageCooldown(DEFAULT_MESSAGE_COOLDOWN);
	if (player !== undefined)
	{
		await runPlayerCommand(player, "title @s actionbar "+message);
		await player.playSound("goe_bos:error", { volume: 0.8 } );
	}
	else
	{
		await runCommand("title @a actionbar "+message);
		await runCommand("playsound goe_bos:error @a ~ ~ ~ 100 1 0.8");
	}
}

export async function showPlayerInfoMessage(player, message)
{
	await setMessageCooldown(DEFAULT_MESSAGE_COOLDOWN);
	if (player !== undefined)
	{
		await runPlayerCommand(player, "title @s actionbar "+message);
		await player.playSound("goe_bos:info", { volume: 0.8 } );
	}
	else
	{
		await runCommand("title @a actionbar "+message);
		await runCommand("playsound goe_bos:info @a ~ ~ ~ 100 1 0.8");
	}
}

export async function tellraw(entity, target, msg) {
	return await entity.runCommandAsync(`tellraw ${target} {"rawtext":[{"text":"§r${msg}"}]}`);
}

export async function title(entity, target, msg) {
	await setMessageCooldown(DEFAULT_MESSAGE_COOLDOWN);
	await runPlayerCommand(entity, `title ${target} times 20 60 20`);
	return await entity.runCommandAsync(`titleraw ${target} title {"rawtext":[{"text":"§r${msg}"}]}`);
}
export async function subtitle(entity, target, msg, addEmptyTitle=false) {
	await setMessageCooldown(DEFAULT_MESSAGE_COOLDOWN);
	await runPlayerCommand(entity, `title ${target} times 20 60 20`);
	await entity.runCommandAsync(`title ${target} subtitle ${msg}`);
	if (addEmptyTitle === true)
		await entity.runCommandAsync(`title ${target} title `);
}
export async function actionbar(entity, target, msg) {
	await setMessageCooldown(DEFAULT_MESSAGE_COOLDOWN);
	await runPlayerCommand(entity, `title ${target} times 20 60 20`);
	return await entity.runCommandAsync(`titleraw ${target} actionbar {"rawtext":[{"text":"§r${msg}"}]}`);
}
export async function titleFull(entity, target, titleMsg, subtitleMsg, actionbarMsg) {
	if (subtitleMsg === undefined)
		subtitleMsg = "§r";
	if (titleMsg === undefined)
		titleMsg = "§r";
	
	await setMessageCooldown(DEFAULT_MESSAGE_COOLDOWN);
	await runPlayerCommand(entity, `title ${target} times 20 60 20`);
	if (actionbarMsg !== undefined)
		await entity.runCommandAsync(`title ${target} actionbar ${actionbarMsg}`);
	await entity.runCommandAsync(`title ${target} title ${titleMsg}`);
	await entity.runCommandAsync(`title ${target} subtitle ${subtitleMsg}`);
}

export async function playSound(entity, target, playSound, volume, pitch, minVolume) {
	let soundVolume = volume ? volume : 100;
	let soundPitch = pitch ? pitch : 1;
	let soundMinVolume = minVolume ? minVolume : 1;
	return await entity.runCommandAsync(`playsound ${playSound} ${target} ~ ~ ~ ${soundVolume} ${soundPitch} ${soundMinVolume}`);
}

export function concat2LinesWithPad(line1, line2)
{
	let len1 = line1.length;
	let len2 = line2.length;
	let message = "";
	if (len1 > len2) 
		message = line1 + "\n" + line2.padStart((len2+((len1-len2)/2)), " ");
	else if (len2 > len1)
		message = line1.padStart((len1+((len2-len1)/2)), " ") + "\n" + line2;
	else 
		message = line1 + "\n" + line2;
	return message;
}
export function debug(message) {
	return world.sendMessage(message);
}

export function typeIdToEntityName(typeId) {
    let name = typeId.split(":")[1]; // Remove prefix:
    name = name.replace(/_/g, ' '); // Replace '_' with ' '
    name = name.split(' ')
        .map(word => word.charAt(0).toUpperCase() + word.slice(1))
        .join(' '); // Replace first letters with uppercase

    return name;
}

export async function setWorldParam(name, value)
{
	world.setDynamicProperty(name, value);
}
export async function loadWorldParam(name, defaultValue)
{
	let param = world.getDynamicProperty(name);
	if (param === undefined)
	{
		param = defaultValue;
		world.setDynamicProperty(name, param);
	}
	return param
}
export async function loadEntityParam(entity, name, defaultValue)
{
	let param = await entity.getDynamicProperty(name);
	if (param === undefined)
	{
		param = defaultValue;
		await entity.setDynamicProperty(name, param);
	}
	return param
}
export async function advanceWorldParam(name)
{
	let param = await world.getDynamicProperty(name);
	param = (param === undefined) ? 1 : param+1;
	await world.setDynamicProperty(name, param);
	return param
}
export async function advanceEntityParam(entity, name)
{
	let param = await entity.getDynamicProperty(name);
	param = (param === undefined) ? 1 : param+1;
	await entity.setDynamicProperty(name, param);
	return param
}

export async function verifyAir(x1, y1, z1, x2, y2, z2)
{
	for (let x=x1; x<=x2; x++)
	{
		for (let y=y1; y<=y2; y++)
		{
			for (let z=z1; z<=z2; z++)
			{
				let block = await overworld.getBlock({x: x, y: y, z: z});
				if (!isBlockAir(block))
					return false;
			}
		}
	}
	return true;
}
export async function verifyAir2(corner1, corner2)
{
	for (let x=corner1.x; x<=corner2.x; x++)
	{
		for (let y=corner1.y; y<=corner2.y; y++)
		{
			for (let z=corner1.z; z<=corner2.z; z++)
			{
				let block = await overworld.getBlock({x: x, y: y, z: z});
				if (!isBlockAir(block))
					return false;
			}
		}
	}
	return true;
}


const ignoreBlocks = [ 'air', 'light_block', 'white_carpet', 'orange_carpet', 'magenta_carpet',
		'light_blue_carpet', 'yellow_carpet', 'lime_carpet', 'pink_carpet', 'gray_carpet', 'light_gray_carpet',
		'cyan_carpet', 'purple_carpet', 'blue_carpet', 'brown_carpet','green_carpet', 'red_carpet', 'black_carpet',
		'grass', 'short_grass', 'tall_grass', 'blue_orchid', 'allium', 'azure_bluet', 'red_tulip', 'orange_tulip', 'white_tulip',
		'pink_tulip', 'oxeye_daisy', 'cornflower', 'lily_of_the_valley', 'wither_rose', 'fern', 'large_fern' ];

export function isBlockAir(block)
{
	if (block === null || block === undefined || block.isAir == undefined || block.isAir) 
		return true;
	if (ignoreBlocks.includes(block.typeId.replace("minecraft:","")))
		return true;
	//if (ignoreBlocks.find((ignoreBlock) => { var _a; return (_a = block) === null || _a === void 0 ? void 0 : _a.permutation.matches(ignoreBlock.block, ignoreBlock.properties); }))
	//	return true;
	return false;
}